# -*- coding: utf-8 -*-

import unittest

from ..fields import ConstantDict

class ConstDemo(ConstantDict):
    TEST = "OK"

class TestMessage(unittest.TestCase):
    def setUp(self):
        pass

    def test_custom_field_getattr(self):
        consttest = ConstDemo()
        self.assertEqual(getattr(consttest, "TEST"), "OK")

    def test_custom_field_missing_getattr(self):
        consttest = ConstDemo()
        try:
            getattr(consttest, "NOTTHERE")
        except AttributeError:
            assert(True)
        else:
            assert(False)


    def test_regular_getattr(self):
        consttest = ConstDemo()
        self.assertIsNotNone(getattr(consttest, "dict"))

    def test_regular_getattr_failure(self):
        consttest = ConstDemo()
        try:
            getattr(consttest, "nowwayxxx")
        except AttributeError:
            assert(True)
        else:
            assert(False)



# -*- coding: utf-8 -*-

import unittest
import logging
import json

from ..taskresultset import TaskResultSet
from ..taskresult import TaskResult
from ..schemas import build_schema
from ..exceptions import TaskResultFailException
from ..fields import MsgErrs as MERR


log = logging.getLogger(name=__name__)


class TestTaskResultSet(unittest.TestCase):
    def setUp(self):
        self.output_name_1 = "test"
        self.task_name_1 = "test_task"

    def test_init(self):
        m = TaskResultSet(4, 5000)
        self.assertIsNotNone(m)

    def test_target_id_job_id_set(self):
        m = TaskResultSet(3, 1000)

        self.assertEqual(m.target_id, 3)
        self.assertEqual(m.job_id, 1000)

    def test_target_id_non_num(self):
        try:
            TaskResultSet("a", 1000)
        except ValueError:
            assert(True)
        else:
            assert(False)

    def test_job_id_non_num(self):
        try:
            TaskResultSet(87, "z")
        except ValueError:
            assert(True)
        else:
            assert(False)

    def test_single_key(self):
        m = TaskResultSet(4, 5000)
        tr = TaskResult(self.output_name_1, 100, self.task_name_1,
                        "1.1.1")
        tr.add("a", "result")
        tr.finish()
        m.add(tr)

        # now extract the test
        tr_extract = m.get_output_tmoid(self.output_name_1, 100)
        self.assertIsNotNone(tr_extract)
        self.assertEqual(tr_extract.success, True)
        self.assertEqual(tr_extract.get_data("a"), "result")

    def test_init_from_dict(self):
        m = TaskResultSet(4, 5000)
        tr = TaskResult(self.output_name_1, 100, self.task_name_1, "1.1.1.1")
        tr.add("a", "result")
        tr.finish()
        m.add(tr)
        logging.debug(m)
        mdict_data = json.loads(json.dumps(m))
        logging.debug(mdict_data)
        mdict = TaskResultSet.init_from_dict(mdict_data)

        self.assertEqual(mdict.target_id, 4)
        self.assertEqual(mdict.job_id, 5000)

        # now extract the test
        tr_extract = mdict.get_output_tmoid(self.output_name_1, 100)
        self.assertIsNotNone(tr_extract)
        self.assertEqual(tr_extract.success, True)
        self.assertEqual(tr_extract.get_data("a"), "result")

    def test_init_from_dict_with_agent(self):
        m = TaskResultSet(4, 5000)
        tr = TaskResult(self.output_name_1, 100, self.task_name_1, "1.1.1.1",
                        agent_name="agent2")
        tr.add("a", "result")
        tr.finish()
        m.add(tr)

        mdict_data = json.loads(json.dumps(m))
        mdict = TaskResultSet.init_from_dict(mdict_data)

        self.assertEqual(mdict.target_id, 4)
        self.assertEqual(mdict.job_id, 5000)

        # now extract the test
        tr_extract = mdict.get_output_tmoid(self.output_name_1, 100)
        self.assertIsNotNone(tr_extract)
        self.assertEqual(tr_extract.success, True)
        self.assertEqual(tr_extract.get_data("a"), "result")
        self.assertEqual(tr_extract.agent_name, "agent2")

    def test_replacement_key(self):
        # check we can overwrite an old result with a new one
        m = TaskResultSet(4, 5000)
        tr = TaskResult(self.output_name_1, 100, self.task_name_1, "1.1.1")
        tr.add("a", "result")
        tr.finish()
        m.add(tr)

        tr2 = TaskResult(self.output_name_1, 100, self.task_name_1, "1.1.1")
        tr2.add("a", "result2")
        tr2.finish()
        m.add(tr2)

        # now extract the test
        tr_extract = m.get_output_tmoid(self.output_name_1, 100)
        self.assertIsNotNone(tr_extract)
        self.assertEqual(tr_extract.success, True)
        self.assertEqual(tr_extract.get_data("a"), "result2")

    def test_missing_task(self):
        m = TaskResultSet(4, 5000)
        try:
            m.get_output_tmoid(self.task_name_1, 100)
        except ValueError:
            assert(True)
        else:
            assert(False)

    def test_missing_tmoid(self):
        m = TaskResultSet(4, 5000)
        tr = TaskResult(self.output_name_1, 100, self.task_name_1, "1.1.1")
        tr.add("a", "result")
        tr.finish()
        m.add(tr)

        try:
            m.get_output_tmoid(self.task_name_1, 999)
        except ValueError:
            assert(True)
        else:
            assert(False)

    def test_unicode(self):
        task = "unicoder_task"
        output = "unicoder"
        key = u"俺"
        val = u"鍵"

        UnicodeSchema = build_schema({
            "properties": {
                "z": {"type": "string"},
                u"俺": {"type": "string"},
            }
        })

        m = TaskResultSet(4, 5000)
        tr = TaskResult(output, 100, task, "1.1.1",
                        schema=UnicodeSchema)
        tr.add(key, val)
        tr.finish()
        m.add(tr)

        tr_extract = m.get_output_tmoid(output, 100, schema=UnicodeSchema)

        self.assertEqual(tr_extract.get_data(key), val)

    # test get_output_tmoid_or_none

    def test_return_none_simple(self):
        m = TaskResultSet(4, 5000)
        tr = TaskResult(self.output_name_1, 100, self.task_name_1, "1.1.1")
        tr.add("a", "result")
        tr.finish()
        m.add(tr)

        ret_tr, err = m.get_output_tmoid_or_none(self.output_name_1, 100)
        print ret_tr, err
        self.assertEqual(ret_tr.get_data("a"), "result")
        self.assertEqual(err, None)

    def test_return_none_on_tr_error(self):
        m = TaskResultSet(4, 5000)
        tr = TaskResult(self.output_name_1, 100, self.task_name_1, "1.1.1")
        tr.add("a", "result")
        tr.finish(err="An error")
        m.add(tr)

        ret_tr, err = m.get_output_tmoid_or_none(self.output_name_1, 100)
        self.assertIsNone(ret_tr)
        self.assertIsNotNone(err)

    def test_return_none_on_no_output(self):
        m = TaskResultSet(4, 5000)
        tr = TaskResult(self.output_name_1, 100, self.task_name_1, "1.1.1")
        tr.add("a", "result")
        tr.finish(err="An error")
        m.add(tr)

        ret_tr, err = m.get_output_tmoid_or_none("not there", 100)
        self.assertIsNone(ret_tr)
        self.assertIsNotNone(err)

    def test_return_none_on_no_tmoid(self):
        m = TaskResultSet(4, 5000)
        tr = TaskResult(self.output_name_1, 100, self.task_name_1, "1.1.1")
        tr.add("a", "result")
        tr.finish(err="An error")
        m.add(tr)

        ret_tr, err = m.get_output_tmoid_or_none("not there", 999)
        self.assertIsNone(ret_tr)
        self.assertIsNotNone(err)

    # test get_output_tmoid_raise_fail

    def test_raise_fail_simple(self):
        m = TaskResultSet(4, 5000)
        tr = TaskResult(self.output_name_1, 100, self.task_name_1, "1.1.1")
        tr.add("a", "result")
        tr.finish()
        m.add(tr)

        ret_tr = m.get_output_tmoid_raise_fail(self.output_name_1, 100)
        self.assertEqual(ret_tr.get_data("a"), "result")

    def test_raise_fail_on_tr_error(self):
        m = TaskResultSet(4, 5000)
        tr = TaskResult(self.output_name_1, 100, self.task_name_1, "1.1.1")
        tr.add("a", "result")
        tr.finish(err="An error")
        m.add(tr)

        try:
            m.get_output_tmoid_raise_fail(self.output_name_1, 100)
        except TaskResultFailException as e:
            self.assertTrue(str(e).startswith(MERR.RESULT_FAIL.format("")))
        else:
            assert(False)

    def test_raise_on_no_output(self):
        m = TaskResultSet(4, 5000)
        tr = TaskResult(self.output_name_1, 100, self.task_name_1, "1.1.1")
        tr.add("a", "result")
        tr.finish(err="An error")
        m.add(tr)

        try:
            m.get_output_tmoid_raise_fail("not there", 100)
        except ValueError:
            assert(True)
        else:
            assert(False)

    def test_raise_on_no_tmoid(self):
        m = TaskResultSet(4, 5000)
        tr = TaskResult(self.output_name_1, 100, self.task_name_1, "1.1.1")
        tr.add("a", "result")
        tr.finish(err="An error")
        m.add(tr)

        try:
            m.get_output_tmoid_raise_fail("not there", 999)
        except ValueError:
            assert(True)
        else:
            assert(False)

    # loops over outputs

    def test_task_item_loop(self):
        TaskLoop1Schema = build_schema({
            "properties": {
                "y": {"type": "string"},
                "z": {"type": ["string", "null"]},
            }
        })

        TaskLoop2Schema = build_schema({
            "properties": {
                "y": {"type": ["string", "null"]},
                "z": {"type": ["string", "null"]},
            }
        })

        m = TaskResultSet(4, 5000)
        tr = TaskResult(
            "output1", 100, "task1", "1.1.1", schema=TaskLoop1Schema)
        tr.add("y", "a")
        tr.add("z", "b")
        tr.finish()

        tr2 = TaskResult(
            "output2", 100, "task2", "1.1.1", schema=TaskLoop2Schema)
        tr2.add("y", "hello")
        tr2.add("z", None)
        tr2.finish()

        m.add(tr)
        m.add(tr2)
        logging.debug(m)
        self.assertEqual(set(m.outputs()), set(["output1", "output2"]))
        self.assertEqual(set(m.output_tmoids("output1")), set([100]))
        self.assertEqual(set(m.output_tmoids("output2")), set([100]))

    def test_task_multi_tmoid(self):
        TaskLoop1Schema = build_schema({
            "properties": {
                "y": {
                    "type": "object",
                    "properties": {
                        "a": {"type": "string"}
                    }
                },
                "z": {"type": ["string", "null"]},
            }
        })

        TaskLoop2Schema = build_schema({
            "properties": {
                "a": {"type": ["string", "null"]},
                "b": {"type": ["string", "null"]},
            }
        })

        m = TaskResultSet(4, 5000)
        tr = TaskResult(
            "output1", 100, "task1", "1.1.1", schema=TaskLoop1Schema)
        tr.add("y", {"a": "1a"})
        tr.add("z", "b")
        tr.finish()

        tr2 = TaskResult(
            "output1", 101, "task1", "1.1.1", schema=TaskLoop1Schema)
        tr2.add("y", {"a": "9z"})
        tr2.add("z", "x")
        tr2.finish()

        tr3 = TaskResult(
            "output2", 100, "task2", "1.1.1", schema=TaskLoop2Schema)
        tr3.add("a", "hello")
        tr3.add("b", "foo")
        tr3.finish()

        m.add(tr)
        m.add(tr2)
        m.add(tr3)

        self.assertEqual(set(m.output_tmoids("output1")), set([100, 101]))
        log.debug(m)
        tr_extract = m.get_output_tmoid("output1", 100, schema=TaskLoop1Schema)
        tr2_extract = m.get_output_tmoid(
            "output1", 101, schema=TaskLoop1Schema)
        tr3_extract = m.get_output_tmoid(
            "output2", 100, schema=TaskLoop2Schema)

        self.assertEqual(tr_extract.get_data_strict("y"), {"a": "1a"})
        self.assertEqual(tr_extract.get_data_strict("z"), "b")
        self.assertEqual(tr2_extract.get_data_strict("y"), {"a": "9z"})
        self.assertEqual(tr2_extract.get_data_strict("z"), "x")
        self.assertEqual(tr3_extract.get_data_strict("a"), "hello")
        self.assertEqual(tr3_extract.get_data_strict("b"), "foo")

    def test_task_level_err(self):
        trs = TaskResultSet(4, 5000)
        trs.add_output_err("whois", "No domains received")

        self.assertEqual(trs.output_errs("whois"), ["No domains received"])

    def test_task_level_success_false(self):
        trs = TaskResultSet(4, 5000)
        trs.add_output_err("whois", "No domains received")

        self.assertFalse(trs.output_success("whois"))

    def test_task_level_success_true(self):
        TaskASchema = build_schema({
            "properties": {
                "a": {"type": "string"}
            }
        })
        print TaskASchema
        trs = TaskResultSet(4, 5000)

        tr = TaskResult("output2", 100, "task2", "1.1.1", schema=TaskASchema)
        tr.add("a", "hello")
        tr.finish()
        trs.add(tr)

        self.assertTrue(trs.output_success("output2"))

    def test_result_array(self):
        IpSchema = build_schema({
            "properties": {
                "ips": {
                    "type": "array",
                    "items": {
                        "type": "string"
                    }
                },
            }
        })

        trs = TaskResultSet(4, 5000)

        ips = TaskResult("output1", 100, "task1", "1.1.1", schema=IpSchema)
        ips.add("ips", ["1.1.1.1", "2.2.2.2"])
        ips.finish()

        trs.add(ips)

        ipset_tmoid = trs.get_output_tmoid("output1", 100, schema=IpSchema)
        ip_tmoid = ipset_tmoid.get_data("ips")

        self.assertEqual(ip_tmoid, ["1.1.1.1", "2.2.2.2"])

    def test_result_nested_obj(self):
        IpSchema = build_schema({
            "properties": {
                "ip": {"type": "string"},
            }
        })

        IpSetSchema = build_schema({
            "properties": {
                "ips": {
                    "type": "object",
                    "patternProperties": {
                        "^(\d{1,3}\.){1,3}(\d{1,3})$": IpSchema
                    },
                    "additionalProperties": False,
                },
            },
            "additionalProperties": False,
        })

        trs = TaskResultSet(4, 5000)
        ipset = TaskResult(
            "output1", 100, "task1", "1.1.1", schema=IpSetSchema)

        ip = TaskResult("output1", 100, "task1", "1.1.1", schema=IpSchema)
        ip.add("ip", "1.1.1.1")
        ip.finish()

        ip2 = TaskResult("output1", 100, "task1", "1.1.1", schema=IpSchema)
        ip2.add("ip", "2.2.2.2")
        ip2.finish()

        ipset.add("ips", {"1.1.1.1": ip, "2.2.2.2": ip2})
        ipset.finish()

        trs.add(ipset)

        ipset_tmoid = trs.get_output_tmoid("output1", 100, schema=IpSetSchema)
        ip_tmoid = ipset_tmoid.get_data("ips").keys()

        self.assertEqual(ip_tmoid, ["1.1.1.1", "2.2.2.2"])



# -*- coding: utf-8 -*-

import logging
import unittest
import json
from freezegun import freeze_time
from datetime import datetime, timedelta
from pytz import utc

from ..taskresult import TaskResult, TaskSeed, SEED_VERSION
from ..fields import MsgFields as F, MsgErrs as MERR
from ..schemas import build_schema


class TestTaskResult(unittest.TestCase):
    def setUp(self):
        self.output_name = "test"
        self.task_name = "test_task"

    def test_init(self):
        m = TaskResult(self.output_name, 111, self.task_name, "1.1.1")
        self.assertIsNotNone(m)

    def test_missing_taskid(self):
        tr = None
        try:
            tr = TaskResult(self.output_name, 111, None, "1.1.1")
        except ValueError:
            assert(True)
        else:
            assert(False)
        self.assertIsNone(tr)

    def test_missing_task_version(self):
        tr = None
        try:
            tr = TaskResult(self.output_name, 111, self.task_name, None)
        except ValueError:
            assert(True)
        else:
            assert(False)
        self.assertIsNone(tr)

    def test_non_str_task_version(self):
        tr = None
        try:
            tr = TaskResult(self.output_name, 111, self.task_name, 1.1)
        except ValueError:
            assert(True)
        else:
            assert(False)
        self.assertIsNone(tr)

    def test_missing_tmoid(self):
        tr = None
        try:
            tr = TaskResult(self.output_name, None, self.task_name, "1.1.1")
        except ValueError:
            assert(True)
        else:
            assert(False)
        self.assertIsNone(tr)

    def test_non_numeric_tmoid(self):
        tr = None
        try:
            tr = TaskResult(self.output_name, "abc", self.task_name, "1.1.1")
        except ValueError:
            assert(True)
        else:
            assert(False)
        self.assertIsNone(tr)

    def test_add(self):
        tr = TaskResult(self.output_name, 38974823, self.task_name, "1.1.1")
        tr.add("a", "Here are my results")
        self.assertEqual(tr.get_data("a"), "Here are my results")

    def test_add_nested_data(self):
        tr = TaskResult(self.output_name, 10098, self.task_name, "1.1.1")
        tr.add("results", {"a": "Nested"})
        self.assertEqual(tr.get_data("results"), {"a": "Nested"})

    def test_missing_data(self):
        tr = TaskResult(self.output_name, 10098, self.task_name, "1.1.1")
        tr.add("a", "Here are my results")
        self.assertIsNone(tr.get_data("nothere"))

    def test_missing_data_strict(self):
        tr = TaskResult(self.output_name, 10098, self.task_name, "1.1.1")
        tr.add("a", "Here are my results")
        try:
            tr.get_data_strict("nothere")
        except ValueError:
            assert(True)
        else:
            assert(False)

    def test_dict_loop(self):
        tr = TaskResult(self.output_name, 8734, self.task_name, "1.1.1")
        tr.add("a", "1")
        tr.add("b", "2")
        self.assertEqual(set(tr.keys()), set(["a", "b"]))

    def test_multi_key(self):
        tr = TaskResult(self.output_name, 8734, self.task_name, "1.1.1")
        tr.add("a", "1")
        tr.add("b", "2")
        tr.finish()

        self.assertEqual(tr.get_data("a"), "1")
        self.assertEqual(tr.get_data("b"), "2")

    def test_finalise_success(self):
        tr = TaskResult(self.output_name, 123, self.task_name, "1.1.1")
        tr.add("a", "result")
        tr.finish()
        self.assertEqual(tr.success, True)

    def test_finalise_failure(self):
        tr = TaskResult(self.output_name, 123, self.task_name, "1.1.1")
        tr.add("a", "result")
        tr.finish(err="Everything broken")
        self.assertFalse(tr.success)
        self.assertIsNotNone(tr.err)

    def test_timer_success(self):
        base_time = datetime(2016, 01, 01, 0, 0, 0)

        with freeze_time(base_time):
            tr = TaskResult(self.output_name, 100, self.task_name, "1.1.1")
            tr.add("a", "My whois data")
        with freeze_time(base_time + timedelta(seconds=2)):
            tr.finish()
        logging.debug(tr)
        self.assertIsNotNone(tr.time_end)
        self.assertLess(tr.time_start, tr.time_end)

    def test_timer_on_error(self):
        base_time = datetime(2016, 01, 01, 0, 0, 0)

        with freeze_time(base_time):
            tr = TaskResult(self.output_name, 100, self.task_name, "1.1.1")
            tr.add("a", "My whois data")
        with freeze_time(base_time + timedelta(seconds=2)):
            tr.finish(err="Everything broke")

        self.assertIsNotNone(tr.time_end)
        self.assertLess(tr.time_start, tr.time_end)

    def test_success(self):
        tr = TaskResult(self.output_name, 100, self.task_name, "1.1.1")
        tr.add("a", "My whois data")
        tr.finish()
        self.assertTrue(tr.success)
        self.assertIsNone(tr.err)

    def test_failure(self):
        tr = TaskResult(self.output_name, 100, self.task_name, "1.1.1")
        tr.add("a", "My whois data")
        tr.finish(err="Everything broke")
        self.assertFalse(tr.success)
        self.assertEqual(tr.err, "Everything broke")

    def test_msg(self):
        tr = TaskResult(self.output_name, 100, self.task_name, "1.1.1")
        tr.add("a", "My whois data")
        tr.finish(msg="Diagnostic info")
        self.assertEqual(tr[F.META][F.META_MSG], "Diagnostic info")

    def test_reserved_name(self):
        tr = TaskResult(self.output_name, 100, self.task_name, "1.1.1")
        try:
            tr.add(F.META, "Reserved field")
        except ValueError:
            assert(True)
        else:
            assert(False)

    def test_version(self):
        tr = TaskResult(self.output_name, 100, self.task_name, "3.4.5")
        tr.add("a", "result")
        tr.finish()

        self.assertEqual(tr.task_version, "3.4.5")

    def test_agent_name_none(self):
        tr = TaskResult(self.output_name, 100, self.task_name, "3.4.5")
        tr.add("a", "result")
        tr.finish()

        self.assertEqual(tr.agent_name, None)

    def test_agent_name(self):
        tr = TaskResult(self.output_name, 100, self.task_name, "3.4.5",
                        agent_name="agent1")
        tr.add("a", "result")
        tr.finish()

        self.assertEqual(tr.agent_name, "agent1")

    def test_obj_json_import(self):
        base_time = datetime(2016, 01, 01, 0, 0, 0)

        with freeze_time(base_time):
            tr = TaskResult(self.output_name, 100, self.task_name, "1.1.1")
            tr.add("a", "My whois data")
            tr.finish(err="Everything broke")

        tr_json = json.dumps(tr)
        logging.debug(tr_json)
        tr2 = TaskResult.init_from_dict(self.output_name,
                                        100, json.loads(tr_json))
        self.assertEqual(tr2.get_data("a"), "My whois data")
        self.assertFalse(tr2.success)
        self.assertEqual(tr2.err, "Everything broke")
        self.assertEqual(tr2.task_name, self.task_name)
        self.assertEqual(tr2.output_name, self.output_name)
        self.assertEqual(tr2.task_version, "1.1.1")
        self.assertEqual(tr2.time_start, base_time.replace(tzinfo=utc))
        self.assertEqual(tr2.time_end, base_time.replace(tzinfo=utc))

    def test_obj_json_import_with_agent(self):
        base_time = datetime(2016, 01, 01, 0, 0, 0)

        with freeze_time(base_time):
            tr = TaskResult(self.output_name, 100, self.task_name, "1.1.1",
                            agent_name="agent1")
            tr.add("a", "My whois data")
            tr.finish()

        tr_json = json.dumps(tr)
        logging.debug(tr_json)
        tr2 = TaskResult.init_from_dict(self.output_name,
                                        100, json.loads(tr_json))
        logging.debug(tr2)
        self.assertEqual(tr2.get_data("a"), "My whois data")
        self.assertTrue(tr2.success)
        self.assertEqual(tr2.err, None)
        self.assertEqual(tr2.agent_name, "agent1")
        self.assertEqual(tr2.task_name, self.task_name)
        self.assertEqual(tr2.output_name, self.output_name)
        self.assertEqual(tr2.task_version, "1.1.1")
        self.assertEqual(tr2.time_start, base_time.replace(tzinfo=utc))
        self.assertEqual(tr2.time_end, base_time.replace(tzinfo=utc))

    def test_schema_fail(self):
        tr = TaskResult(self.output_name, 100, self.task_name, "1.1.1")
        tr.add("nosuch", "My whois data")
        tr.finish()

        self.assertFalse(tr.success)
        self.assertTrue(tr.err.startswith(MERR.RESULT_INVALID.format("")))

    def test_custom_schema(self):
        TestCustomSchema = build_schema({
            "properties": {
                "z": {"type": "string"}
            },
            "required": ["z"]
        })

        tr = TaskResult(self.output_name, 100, self.task_name, "1.1.1",
                        schema=TestCustomSchema)
        tr.add("z", "My whois data")
        tr.finish()

        self.assertEqual(tr.get_data("z"), "My whois data")

    def test_custom_schema_invalid(self):
        TestCustomInvalidSchema = build_schema({
            "properties": {
                "z": {"type": "string"}
            },
            "required": ["z"]
        })

        tr = TaskResult(self.output_name, 100, self.task_name, "1.1.1",
                        schema=TestCustomInvalidSchema)
        tr.add("y", "My whois data")
        tr.finish()

        self.assertFalse(tr.success)
        self.assertTrue(tr.err.startswith(MERR.RESULT_INVALID.format("")))


class TestTaskSeed(unittest.TestCase):
    def setUp(self):
        self.output_name = "test"

    def test_taskseed_payload(self):
        ts = TaskSeed(self.output_name, 100, "a", "example.com")
        self.assertEqual(ts.get_data("a"), "example.com")

    def test_taskseed_task_name(self):
        ts = TaskSeed(self.output_name, 100, "a", "example.com")
        self.assertEqual(ts.task_name, F.SEEDED_TASK)

    def test_taskseed_version(self):
        ts = TaskSeed(self.output_name, 100, "a", "example.com")
        self.assertEqual(ts.task_version, SEED_VERSION)

    def test_taskseed_dates(self):
        ts = TaskSeed(self.output_name, 100, "a", "example.com")
        self.assertIsNotNone(ts.get_meta(F.META_START_TIME))
        self.assertIsNotNone(ts.get_meta(F.META_END_TIME))



# -*- coding: utf-8 -*-

import logging
import jsonschema
import iso8601

from datetime import datetime
from .fields import MsgFields as F, MsgErrs as MERR
from .exceptions import ValidationError
import schemas


log = logging.getLogger(__name__)

SEED_VERSION = "9.9.9"


def iso_now():
    return datetime.utcnow().isoformat()


class TaskResult(dict):
    """
    TaskResult represents a Task result for a single Target Monitoring Object
    (TMO). It's designed to be used inside task functions like this:

    tr = TaskResult("mytask", "2.2.5", 1111)
    tr.add("analysis", [1, 2, 3])
    tr.finish()

    This also adds a variety of metadata, like task run times and errors.
    """

    reserved_keys = [F.META]

    def __init__(self, output_name, tmoid, task_name, task_version,
                 agent_name=None, data=None, schema=None):
        # str name of output type this TaskResult generated Not used
        # internally, but referenced when adding to TaskResults.
        self.output_name = output_name
        # int with the id of the target monitor object Not used internally, but
        # referenced when adding to TaskResults.
        self.tmoid = tmoid
        # agent that fulfiled this request
        self.__agent_name = agent_name
        # str name of the task generating this output.
        self.__task_name = task_name
        # Supplied by task.
        self.__task_version = task_version

        # Should be class of type colander.Schema.
        # Chiefly for testing - by default, will look for corresponding
        # schema to task_name
        self.schema = self._set_schema(schema)

        if not self.output_name:
            raise ValueError("Must define an output type")
        if not self.__task_name:
            raise ValueError("Must define a task type")
        if (not self.__task_version or
                not isinstance(self.__task_version, basestring)):
            raise ValueError("Must define a task version as a string "
                             "(e.g. '1.2.2.'")

        try:
            self.tmoid = str(int(self.tmoid))
        except TypeError:
            raise ValueError("Target monitoring id (tmoid) must be an int: "
                             "got %s of type %s"
                             % (self.tmoid, type(self.tmoid)))

        # start the timer
        self._time_start = iso_now()
        self._time_end = None

        dict.__init__(self)

        if data and not isinstance(data, dict):
            raise ValueError("data must be a dictionary to import")
        elif data:
            self.update(data)

    @classmethod
    def init_from_dict(cls, output_name, tmoid, data, schema=None):
        """
        Load a TaskResult from a dict, likely a deserialized TaskResult from
        somewhere else. Note that task_version is contained in the meta data
        already, so we don't pass it in.
        """
        # str name of the task's output
        task_version = None
        # int with the id of the target monitor object

        try:
            task_name = data[F.META][F.META_TASK_NAME]
            task_version = data[F.META][F.META_TASK_VERSION]
            agent_name = data[F.META].get(F.META_AGENT_NAME)
        except KeyError:
            raise ValueError("TaskResult dict does not define a version "
                             "number")
        return cls(output_name, tmoid, task_name, task_version,
                   agent_name=agent_name, data=data, schema=schema)

    def _set_schema(self, supplied_schema):
        """
        Automatically select a JSON schema to validate this TaskResult.
        This is either based on the task_name (e.g. if task_name is abc,
        _set_schema will attempt to load AbcSchema, or load a schema manually
        supplied).
        """
        schema_cls = None

        if not self.output_name:
            raise ValueError("Must specify a task_name")

        if not supplied_schema:
            log.debug("Output name: {0}".format(self.output_name))
            log.debug("Output name title: {0}"
                      .format(self.output_name.title()))
            schema_name = "{0}Schema".format(self.output_name.title())
            try:
                schema_cls = getattr(schemas, schema_name)
            except AttributeError:
                raise ValueError("Couldn't find validation schema called {0}"
                                 .format(schema_name))
        else:
            schema_cls = supplied_schema

        return schema_cls

    def add(self, result_name, result_val):
        """
        Add a named result to the TaskResult.
        """
        if result_name in self.reserved_keys:
            raise ValueError("{0} is a reserved key name - please choose"
                             " another.".format(result_name))

        if not result_name:
            raise ValueError("key must be a defined value")

        self.setdefault(result_name, result_val)

    def get_data_strict(self, key):
        """
        Find and return a data key, raise ValueError if not found
        """
        try:
            return self[key]
        except KeyError:
            log.debug(self)
            raise ValueError("{0}.{1} has no value for {2}".format(
                self.output_name, self.tmoid, key))

    def get_data(self, key):
        """
        Find and return a data key, return None if not found
        """
        return self.get(key, None)

    def get_meta(self, key):
        """
        Return a meta field from this Task
        """
        return self[F.META].get(key)

    @property
    def success(self):
        """
        Return true if the task returned with no errors
        """
        return self.err is None

    @property
    def err(self):
        """
        Return errors returned by a task
        """
        try:
            return self[F.META][F.META_ERR]
        except KeyError:
            return

    @property
    def msg(self):
        """
        Return diagnostic messages returned by a task
        """
        try:
            return self[F.META][F.META_MSG]
        except KeyError:
            return

    @property
    def agent_name(self):
        """
        Return name of agent executing task
        """
        try:
            return self[F.META][F.META_AGENT_NAME]
        except KeyError:
            return

    @property
    def task_name(self):
        """
        Return task name that executed
        """
        try:
            return self[F.META][F.META_TASK_NAME]
        except KeyError:
            return

    @property
    def task_version(self):
        """
        Return errors returned by a task
        """
        try:
            return self[F.META][F.META_TASK_VERSION]
        except KeyError:
            return

    @property
    def time_start_iso(self):
        """
        Return start time for a task
        """
        try:
            return self[F.META][F.META_START_TIME]
        except KeyError:
            return

    @property
    def time_start(self):
        return iso8601.parse_date(self.time_start_iso)

    @property
    def time_end_iso(self):
        """
        Return start time for a task
        """
        try:
            return self[F.META][F.META_END_TIME]
        except KeyError:
            return

    @property
    def time_end(self):
        return iso8601.parse_date(self.time_end_iso)

    def finish(self, err=None, msg=None):
        """
        Add the result for the event, any error messages, and stop the timer
        """
        self._time_end = iso_now()

        self.setdefault(F.META, {})
        self[F.META][F.META_ERR] = err
        self[F.META][F.META_MSG] = msg
        self[F.META][F.META_START_TIME] = self._time_start
        self[F.META][F.META_END_TIME] = self._time_end
        self[F.META][F.META_AGENT_NAME] = self.__agent_name
        self[F.META][F.META_TASK_NAME] = self.__task_name
        self[F.META][F.META_TASK_VERSION] = self.__task_version

        try:
            self.schema_validate()
        except ValidationError as e:
            # The original error may have caused schema non-validation.
            # Leave original error if exists, otherwise, add schema err.
            if not self[F.META][F.META_ERR]:
                self[F.META][F.META_ERR] = MERR.RESULT_INVALID.format(e)

        if self.err:
            log.debug(self.err)

    def schema_validate(self):
        try:
            jsonschema.validate(self, self.schema)
        except Exception as e:
            log.debug(self)
            raise ValidationError(
                "Couldn't validate under {0} schema: {1}"
                .format(self, e)
            )


class TaskSeed(TaskResult):
    def __init__(self, output_name, tmoid, key, val):
        super(TaskSeed, self).__init__(output_name, tmoid, F.SEEDED_TASK,
                                       SEED_VERSION)
        self.add(key, val)
        self.finish()
